//------------------------------------------------------------------------------
// Прошивка для твердотельного реле
// Код семейства - 0xA2
// Адрес на шине i2c - 0x12
// OldBean, 28.11.2018, 22.07.2019
//------------------------------------------------------------------------------
#include <avr/io.h>
#include <util/twi.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <string.h>
//------------------------------------------------------------------------------
#define F_CPU 16000000UL // Тактовая частота МК
#include <util/delay.h>
//------------------------------------------------------------------------------
#define ADDR 0x12        // Адрес устройства на шине I2C
#define FAMILY_CODE 0xA2 // Код семейства данного устройства
#define BUF_SIZE 32      // Размер буфера обмена данными
#define EE_SIZE 6        // Размер блока идентиф. и калибр. данных
//------------------------------------------------------------------------------
// Команды:
// 0x33 Чтение кода семейства
// 0x4e Записать данные из буфера обмена в рабочую область (write scratch)
// 0xbe Считать данные из рабочей области в буфер обмена (read scratch)
// 0x48 Записать блок данных из рабочей области в EEPROM (copy scratch)
// 0xb8 Считать блок данных из EEPROM в рабочую область (recall EEPROM)
// 0x52 Запись данных в рабочие регистры (для контроллеров)
//------------------------------------------------------------------------------
static uint8_t buf[BUF_SIZE];        // Буфер обмена
static uint8_t scratch[BUF_SIZE];    // Рабочая область RAM
static volatile uint8_t *ptr;        // Указатель на байт буфера обмена
static volatile uint8_t runFlag = 1; // Рабочий режим, 0 - режим настройки
static volatile uint8_t nb = 0;      // Счетчик байтов
static volatile uint8_t prFlag = 0;  // Принят новый пакет байтов
static volatile uint8_t cmd = 0x00;  // Код команды
//------------------------------------------------------------------------------
ISR(TWI_vect) { // Обработчик событий модуля TWI
  switch(TW_STATUS) {     // Статус - в 5 старших битах регистра TWSR
    // Принимаем байты
    case TW_SR_SLA_ACK:   // Адрес для приема принят, ACK отправлен - 0x80
      ptr = buf;          // Настраиваемся на начало буфера обмена
      nb = 0;             // Сброс счетчика байтов
      break;
    case TW_SR_DATA_ACK:  // Очередной байт принят, ACK отправлен - 0x80
      *ptr++ = TWDR;
      nb++;
      break;
    case TW_SR_STOP:      // Состояние "стоп" на шине - 0xA0
      prFlag = 1;         // Новый пакет принят
      break;
    // Отправляем байты
    case TW_ST_SLA_ACK:   // Адрес для отправки принят, ACK отправлен - 0xA8
      ptr = buf;          // Настраиваемся на начало буфера обмена
      nb = 0;             // Сброс счетчика байтов
    case TW_ST_DATA_ACK:  // Очередной байт отправлен, ACK принят - 0xB8
      TWDR = *ptr++;
      nb++;
      break;
    case TW_ST_DATA_NACK: // Байт отправлен и получен NACK - 0xC0
    case TW_ST_LAST_DATA: // Отправлен последний байт, получен ACK - 0xC8
      break;
    // Ошибки старта или стопа - сброс TWI
    case TW_BUS_ERROR:
      TWCR = _BV(TWIE) | _BV(TWINT) | _BV(TWEA) | _BV(TWEN);
      break;
  }
  TWCR |= _BV(TWINT); // Очистим флаг TWINT (очищается при установке в 1)
}
//------------------------------------------------------------------------------
ISR(INT1_vect) { // Внешнее прерывание INT1 - авария
  PORTD &= ~_BV(PD0); // Выключаем реле
}
//------------------------------------------------------------------------------
int main(void) {
  DDRD |= _BV(PD0);   // Пин 0 порта D - на вывод
  // Инициализация TWI-модуля --------------------------------------------------
  TWAR = ADDR << 1;   // Адрес устройства - в старших 7 битах
  TWCR = _BV(TWIE);   // Разрешаем прерывания от TWI
  TWCR |= _BV(TWEA);  // Разрешаем подачу импульса ACK
  TWCR |= _BV(TWINT); // Очистим флаг TWINT (очищается при установке в 1)
  TWCR |= _BV(TWEN);  // Разрешаем работу TWI и активирует TWI-интерфейс
  // Инициализация внешнего прерывания INT1
  EICRA |= _BV(ISC11); // Добавляем прерывания по спаду уровня на пине PD3/INT1
  EIMSK |= _BV(INT1); // Добавляем разрешение на внешние прерывания INT1

  sei(); // Разрешаем глобальные прерывания

  ptr = buf;
  for(int i = 0; i < BUF_SIZE; buf[i++] = 15);
  while(1) { // Бесконечный цикл
    if(prFlag) { // По шине i2c принят новый пакет - отрабатываем его
      TWCR &= ~_BV(TWEN); // Приостановим работу TWI
      prFlag = 0; // Сбросим флаг получения нового пакета
      cmd = buf[0]; // Код команды
      switch(cmd) { // Анализируем и выполняем команды
        case 0x33: // Чтение ROM - в первом байте - код семейства
          buf[0] = FAMILY_CODE;
          break;
        case 0x4e: // write scratch - записать данные в рабочую область
          memcpy(scratch, &buf[1], EE_SIZE);
          break;
        case 0xbe: // read scratch - считать данные из рабочей области
          memcpy(buf, scratch, EE_SIZE);
          break;
        case 0x48: // copy scratch - записать блок данных в EEPROM
          eeprom_update_block((const void *)scratch, (void *)0, EE_SIZE);
          break;
        case 0xb8: // recall EEPROM - считать блок данных в EEPROM
          eeprom_read_block((void *)scratch, (const void*)0, EE_SIZE);
          break;
        case 0x52: // Команда записи данных в регистры контроллеров
          if(buf[1]) PORTD |= _BV(PD0); // Не ноль - включаем
          else PORTD &= ~_BV(PD0); // Если 0, то выключаем
          break;
      }
      TWCR |= _BV(TWEN); // Активируем TWI
    }
  }
  return 0;
}
//------------------------------------------------------------------------------
// История и список изменений
// 1. 28.11.2018 - Отлажен весь функционал для этого типа устройств.
// 2. 26.07.2019 - Добавлена блокировка TWI на время обработки команд
// 3. 16.08.2019 - Добавлена обработка прерывания INT1 - авария
//------------------------------------------------------------------------------
