//------------------------------------------------------------------------------
// Прошивка для 10-разрядного АЦП с 1-Wire интерфейсом. После старта
// преобразования (код 0x44) производится усреднение по 4096 выборкам.
// Протокол 1-Wire упрощен (только одно устройство на линии и отсутствие CRC)
// OldBean, 30.11.2018
//------------------------------------------------------------------------------
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <string.h>
//------------------------------------------------------------------------------
#define F_CPU 16000000UL // Тактовая частота МК
#include <util/delay.h>

#define FAMILY_CODE 0xA6 // Код семейства данного устройства
#define BUF_SIZE 32      // Размер буфера обмена данными
#define EE_SIZE 6        // Размер блока идентиф. и калибр. данных
//------------------------------------------------------------------------------
static uint16_t adcv;                // Код АЦП, передаваемый мастеру шины
static uint32_t sum;                 // Сумматор для усреднения
static uint16_t cntr;                // Счетчик просуммированных выборок
volatile static uint8_t adcFlag = 0; // Если 1, то измерения разрешены
//------------------------------------------------------------------------------
static uint8_t buf[BUF_SIZE];            // Буфер обмена
static volatile uint8_t dataFlag = 0;    // Если 0 - ожидается байт с командой,
                                         // если 1 - то данные
static volatile uint8_t srFlag = 0;      // skip ROM флаг
static volatile uint8_t *ptr;            // Указатель на байт буфера обмена
static volatile uint8_t nb = 0;          // Кол-во байтов данных для обмена
static volatile uint8_t command = 0x00;  // Код команды
//------------------------------------------------------------------------------
volatile static uint8_t wFlag = 0;   // 1 - slave пишет в линию, 0 - читает
//volatile static uint8_t srFlag = 0;  // skip ROM получен - работаем
volatile static uint8_t mask = 0x01; // Маска для "разборки" байта на биты
volatile static uint8_t data = 0;    // Байт принятых/отправляемых данных
static uint16_t tau; // Длительность низкого уровня после старта тайм-слота
//------------------------------------------------------------------------------
ISR(PCINT0_vect){ // Обработчик внешнего прерывания на PB3
  if(!(PINB & _BV(PB3))) { // Был переход линии с высокого уровня на низкий
    GTCCR = _BV(PSR0); // Запускаем таймер: сброс предделителя таймера 0
    TCNT0 = 0;         // и обнуляем счетчик таймера 0
    if(wFlag) { // Ведомый передает биты
      if(!(data & mask)) { // Нужно передать нулевой бит
        PORTB &= ~_BV(PB3); // Устанавливаем низкий уровень на пине PB3
        DDRB |= _BV(PB3); // Перевод пина PB3 в режим записи (прижмем линию)
        _delay_us(20); // Держим паузу (не менее 15 мкс)
        DDRB &= ~_BV(PB3); // Перевод пина PB3 в режим чтения (отпустим линию)
      } // А единичный бит получится автоматически, если ничего не делать ;)
      mask <<= 1; // Модифицируем маску (переходим на следующий бит)
      if(mask == 0) { // Отправлен целый байт
        if(--nb == 0) { // Ну, наконец, все байты отправлены
          wFlag = 0; // Переходим в режим чтения
          dataFlag = 0; // Сбрасываем флаг приема данных
          command = 0x00;
        } else { // Продолжаем отправлять байты
          data = *ptr++; // Следующий байт к отправке
        }
        mask = 0x01; // Сбрасываем маску на младший бит
      }
    }
  } else { // И наоборот - это переход линии с низкого на высокий
    // Оценим длительность импульса (низкого уровня на линии) tau в мкс
    tau = TCNT0; tau <<= 2; // Единица счетчика TIMER0 - это 4 мкс
    if(tau >= 480) { // Это RESET. Для простоты импульс PRESENCE пока не подаем
      wFlag = 0; // Переходим в режим чтения
      mask = 0x01; // Сбрасываем маску на младший бит
      data = 0; // Очищаем принимаемый байт
      dataFlag = 0; // После сброса всегда ждем команду...
      srFlag = 0; // Очищаем флаг skip ROM
    } else { // Слот или кусок слота
      if(!wFlag) { // Режим чтения
        if(tau < 15) { // На линии высокий уровень - это единичный бит
          data |= mask; // Примем его
        } // А нулевой бит будет считан автоматически, если ничего не делать
        mask <<= 1; // Модифицируем маску (переходим на следующий бит)
        if(mask == 0) { // Принят целый байт
          if(dataFlag) { // Принимаем данные (не команды)
            *ptr++ = data; // Заносим принятый байт в буфер обмена
            data = 0;
            if(--nb == 0) { // Ну, наконец, все байты приняты
              dataFlag = 0; // Сбрасываем флаг приема данных
              command = 0x00;
            }
          } else { // Принята команда
            command = data;
            switch(command) {
              case 0x33: // read ROM - код семейства будет в первом байте
                data = FAMILY_CODE; // Байт к отправке
                nb = 1;        // Отдадим всего один байт
                wFlag = 1;     // Направление передачи - slave отправляет байты
                dataFlag = 1;  // Режим приема/передачи данных
                break;
              case 0xcc: // Получен skip ROM  - !!! Пока не обрабатывается !!!
                srFlag = 1;    // Установим флаг
                wFlag = 0;     // Направление передачи - slave получает байты
                data = 0;      // Очищаем принимаемый байт
                break;
              case 0x44: // Получена команда - запуск преобразования
                sum = 0;       // Очищаем сумматор
                cntr = 0;      // Сброс счетчика выборок
                adcFlag = 1;   // Разрешаем измерение и суммирование
                data = 0;      // Очищаем принимаемый байт
                break;
              case 0x4e: // Принимаем блок данных от мастера
                ptr = buf + 2; // Указатель - на начало буфера обмена + 2 
                data = 0;      // Очищаем принимаемый байт
                nb = EE_SIZE;  // Размер блока данных
                wFlag = 0;     // Направление передачи - slave получает байты
                dataFlag = 1;  // Режим приема/передачи данных
                break;
              case 0xbe: // Отправляем блок данных мастеру
                ptr = buf;     // Указатель - на начало буфера обмена
                data = *ptr++;
                nb = EE_SIZE + 2;  // Размер блока данных
                wFlag = 1;     // Направление передачи - slave отправляет байты
                dataFlag = 1;  // Режим приема/передачи данных
                break;
              default:         // Неизвестная команда
                wFlag = 0;     // Направление передачи - slave получает байты
                data = 0;      // Очищаем принимаемый байт
                break;
            }
          }
          mask = 0x01; // Сбрасываем маску на младший бит
        }
      }
    }
  }
}
//------------------------------------------------------------------------------
int main(void) {
//  MCUCR |= _BV(PUD);  // Отключаем подтягивающие резисторы на всех портах
//  DDRB = _BV(PB0) | _BV(PB1) | _BV(PB2); // Пины для диагностики
  // Инициализация внешних прерываний на пине PB3 ------------------------------
  GIMSK = _BV(PCIE);   // Прерывания по изменению уровня на пине
  PCMSK = _BV(PCINT3); // Внешние прерывания ловим на пине PB3
  // Инициализация таймера 0 для тактирования АЦП ------------------------------
  // Счетчик работает в нормальном режиме
  TCCR0B = _BV(CS01) | _BV(CS00); // Предделитель на 64
  // Инициализация АЦП ---------------------------------------------------------
  ADMUX = 0x00; // В качестве источника опорного напряжения - питание
  ADMUX |= _BV(MUX1); // Работаем в ADC2 (источник подключен к пину PB4)
  ADCSRA = _BV(ADEN); // АЦП включен
  ADCSRA |= _BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0); // Предделитель на 128

  sei(); // Разрешаем глобальные прерывания
  while(1) { // Бесконечный цикл
    switch(command) {
      case 0x48: // Записать блок байтов из буфера обмена в EEPROM
        eeprom_update_block((const void *)(buf + 2), (void *)0, EE_SIZE);
        wFlag = 0;     // Направление передачи - slave получает байты
        data = 0;      // Очищаем принимаемый байт
        break;
      case 0xb8: // Считать блок байтов из EEPROM в буфер обмена
        eeprom_read_block((void *)(buf + 2), (const void*)0, EE_SIZE);
        wFlag = 0;     // Направление передачи - slave получает байты
        data = 0;      // Очищаем принимаемый байт
        break;
    }

    if(!adcFlag) continue; // Измерения запрещены
    ADCSRA |= _BV(ADSC); // Запуск АЦП
    while(ADCSRA & _BV(ADSC)); // Ждем пока завершится преобразование
    // Часть операций ниже не являются атомарными, но в периоды времени,
    // когда они выполняются никаких прерываний не инициируется, т.к.
    // мастер делает паузу примерно на 750 мс и не "дергает" шину.
    // В связи с этим никаких дополнительных мер по безопасности следующего
    // ниже кода не применяется
    sum += ADCL;
    sum += (ADCH << 8);
    if(++cntr == 4096) { // Достаточно насуммировали...
      sum += 2048; // Ну усреднять, так усреднять ;) 
      adcv = sum >> 12; // Усредненный по 4096 выборкам код АЦП
      adcFlag = 0; // Дальнейшие измерения до следующей команды 0x44 запрещаем
      buf[0] = adcv & 0xff;
      buf[1] = (adcv >> 8) & 0xff;
    }
  }
  return 0;
}
//------------------------------------------------------------------------------
// История и список изменений
// 1. 30.11.2018 - Отлажен базовый функционал для этого типа устройств.
//------------------------------------------------------------------------------
